"""OAuth2 session client to handle authentication of calls to the Monzo API

Adapted from the implementation for a Fitbit API Python client by Orcas.inc

Used under the Apache2 license: https://github.com/orcasgit/python-fitbit

"""

from typing import Dict, Tuple, Callable
import json
import requests

from requests.auth import HTTPBasicAuth
from requests_oauthlib import OAuth2Session

from monzo.errors import (BadRequestError, UnauthorizedError, ForbiddenError,
MethodNotAllowedError, PageNotFoundError, NotAcceptibleError,TooManyRequestsError,
InternalServerError, GatewayTimeoutError)

def print_token(token:Dict[str,str])-> Dict[str,str]:
    print(token)

class MonzoOauth2Client(object):
    AUTHORIZE_ENDPOINT = "https://auth.monzo.com"
    API_ENDPOINT = "https://api.monzo.com"
    API_VERSION = 1

    authorization_url = AUTHORIZE_ENDPOINT
    request_token_url = "%s/oauth2/token" % API_ENDPOINT
    access_token_url = request_token_url
    refresh_token_url = request_token_url

    def __init__(self, client_id:str, client_secret:str, access_token:str=None,
            refresh_token:str=None, expires_at:float=None, refresh_cb:Callable[[str],None]=None,
            redirect_uri:str=None, *args, **kwargs):
        """
        Create a MonzoOauth2Client object. Specify the first 7 parameters if
        you have them to access user data. Specify just the first 2 parameters
        to start the setup for user authorization (as an example see gather_key_oauth2.py)
            - client_id, client_secret are in the app configuration page
            https://developers.monzo.com/
            - access_token, refresh_token are obtained after the user grants permission
        """

        self.client_id, self.client_secret = client_id, client_secret
        token = {}
        if access_token:
            token.update({'access_token': access_token})
        if refresh_token:
            token.update({'refresh_token': refresh_token})
        if expires_at:
            token['expires_at'] = expires_at

        self.session = OAuth2Session(
            client_id,
            auto_refresh_url=self.refresh_token_url,
            token_updater=refresh_cb,
            token=token,
            redirect_uri=redirect_uri,
        )
        self.timeout = kwargs.get("timeout", None)

    def _request(self, method:str, url:str, **kwargs):
        """
        A simple wrapper around requests.
        """
        if self.timeout is not None and 'timeout' not in kwargs:
            kwargs['timeout'] = self.timeout

        try:
            response = self.session.request(method, url, **kwargs)

            # If our current token has no expires_at, or something manages to slip
            # through that check
            if response.status_code == 401:
                d = json.loads(response.content.decode('utf8'))
                if d['errors'][0]['errorType'] == 'expired_token':
                    self.refresh_token()
                    response = self.session.request(method, url, **kwargs)

            return response
        except requests.Timeout as e:
            pass
            #raise exceptions.Timeout(*e.args)

    def make_request(self, url:str, data:Dict=None, method:str=None, **kwargs):
        """
        Builds and makes the OAuth2 Request, catches errors
        https://docs.monzo.com/#errors
        """
        data = data or {}
        method = method or ('POST' if data else 'GET')
        response = self._request(
            method,
            url,
            data=data,
            client_id=self.client_id,
            client_secret=self.client_secret,
            **kwargs
        )

        return self.validate_response(response)

    def authorize_token_url(self, redirect_uri:str = None, **kwargs) -> Tuple[str,str]:
        """Step 1: Return the URL the user needs to go to in order to grant us
        authorization to look at their data.  Then redirect the user to that
        URL, open their browser to it, or tell them to copy the URL into their
        browser.

            :param redirect_uri: url to which the response will posted.
            :rtype: A Tuple consisting of the authentication url and the state token.
        """

        if redirect_uri:
            self.session.redirect_uri = redirect_uri

        return(self.session.authorization_url(self.authorization_url, **kwargs))

    def fetch_access_token(self, code:str, redirect_uri:str=None):
        """Step 2: Given the code from fitbit from step 1, call
        fitbit again and returns an access token object. Extract the needed
        information from that and save it to use in future API calls.
        the token is internally saved

            :rtype: A Dictionary representation of the authentication status.
        """
        if redirect_uri:
            self.session.redirect_uri = redirect_uri
        return self.session.fetch_token(
            self.access_token_url,
            username=self.client_id,
            password=self.client_secret,
            code=code)

    def refresh_token(self) -> Dict[str,str]:
        """Step 3: obtains a new access_token from the the refresh token
        obtained in step 2. Only do the refresh if there is `token_updater(),`
        which saves the token.

            :rtype: A Dictionary representation of the authentication token.
        """
        token = {}
        if self.session.token_updater:
            token = self.session.refresh_token(
                self.refresh_token_url,
                auth=HTTPBasicAuth(self.client_id, self.client_secret)
            )
            self.session.token_updater(token)

        return token

    def validate_response(self, response):
        """Validate the response and raises any appropriate errors.
           https://docs.monzo.com/#errors

           :param response: The response to validate
           :rtype: A Dictionary representation of the response, if no errors occured.
        """
        json_response = response.json()
        if response.status_code == 200:
            return json_response
        if response.status_code == 400:
            raise BadRequestError(json_response['message'])
        if response.status_code == 401:
            raise UnauthorizedError(json_response['message'])
        if response.status_code == 403:
            raise ForbiddenError(json_response['message'])
        if response.status_code == 404:
            raise PageNotFoundError(json_response['message'])
        if response.status_code == 405:
            raise MethodNotAllowedError(json_response['message'])
        if response.status_code == 406:
            raise NotAcceptibleError(json_response['message'])
        if response.status_code == 429:
            raise TooManyRequestsError(json_response['message'])
        if response.status_code == 500:
            raise InternalServerError(json_response['message'])
        if response.status_code == 504:
            raise GatewayTimeoutError(json_response['message'])



"""
The module that represents each possible Monzo API error; documented at:
https://monzo.co.uk/docs/#errors as individual Python errors.
"""
class BadRequestError(Exception):
    """An error to be raised when a request has missing arguments or is malformed."""

class UnauthorizedError(Exception):
    """An error to be raised when a request is not authenticated."""

class ForbiddenError(Exception):
    """An error to be raised when a request has insufficient permissions."""

class MethodNotAllowedError(Exception):
    """An error to be raised when a request is using an incorrect HTTP verb."""

class PageNotFoundError(Exception):
    """An error to be raised when a request requests an endpoint that doesn't exist."""

class NotAcceptibleError(Exception):
    """An error to be raised when an application does not accept the content
    format returned according to the Accept headers sent in the request."""

class TooManyRequestsError(Exception):
    """An error to be raised when an application is exceeding its rate limit.
    Back off, buddy. :p"""

class InternalServerError(Exception):
    """An error with Monzo's servers."""

class GatewayTimeoutError(Exception):
    """A timeout has occured on Monzo's servers."""



"""A wrapper around the official Monzo API endpoints.

This module contains the class `Monzo` which represents a wrapper around
HTTP calls to Monzo's API endpoints.
"""

from typing import Dict, List, Union, Any
import monzo.auth as _auth
import string
import random
import webbrowser

class Monzo(object):
    """The class representation of Monzo's API endpoints.

       Please note that functions without a reference to the official Monzo API
       docs page are convenience functions which are created - based on the official
       API functions - to make life easier for developers.

       e.g. `get_first_account` calls `get_account` and returns the first `account`
       object, if it exists.

       :param access_token: The access token to authorise API calls.
    """

    API_URL = 'https://api.monzo.com/' #: (str): A representation of the current Monzo api url.

    def __init__(self, client_id:str = None, client_secret:str= None, access_token:str = None,
                 refresh_token:str = None, expires_at:str = None, refresh_cb=lambda x: None) -> None:
        self.oauth_session = _auth.MonzoOauth2Client(client_id, client_secret, access_token=access_token, refresh_cb=refresh_cb)
        #if not access_token:
        #    webbrowser.open(self.oauth_session.authorize_token_url()[0], new=2)
        #    auth_code = input("Input Auth code from email")
        #    self.oauth_session.fetch_access_token(auth_code)
        #    self.access_token = self.oauth_session.session.token

    def update_token(self,token:Dict[str,str]) -> None:
        """Updates locally stored access token

        Called by MonzoOauth2Client.refresh_token() when previous token expires

           :param token: The updated Oauth token generated by MonzoOauth2Client.
        """
        self.access_token=token['access_token']

    def whoami(self) -> Dict[str,str]:
        """Gives information about an access token. (https://monzo.com/docs/#authenticating-requests)

           :rtype: A Dictionary representation of the authentication status.
        """
        url = "{0}/ping/whoami".format(self.API_URL)
        response = self.oauth_session.make_request(url)
        return response

    def get_accounts(self) -> Dict[str,List[Dict[str,str]]]:
        """Get all accounts that belong to a user. (https://monzo.com/docs/#list-accounts)

           :rtype: A Collection of accounts for a user.

        """
        url = "{0}/accounts".format(self.API_URL)
        response = self.oauth_session.make_request(url)
        return response

    def get_first_account(self) -> Dict[str,Union[str,List]]:
        """Gets the first account for a user.

           :rtype: A Dictionary representation of the first account belonging to a user, if it exists.

        """
        accounts = self.get_accounts()
        if len(accounts['accounts']) <= 0:
            raise LookupError('There are no accounts associated with this user.')
        return accounts['accounts'][0]

    def get_transactions(self, account_id) -> Dict[str,List[Dict]]:
        """Get all transactions of a given account. (https://monzo.com/docs/#list-transactions)

           :param account_id: The unique identifier for the account which the transactions belong to.
           :rtype: A collection of transaction objects for specific user.
        """
        url = "{0}/transactions".format(self.API_URL)
        params = {'expand[]': 'merchant', 'account_id': account_id}
        response = self.oauth_session.make_request(url, params=params)
        return response

    def get_balance(self, account_id) -> Dict[str, Any]:
        """Gets the balance of a given account. (https://monzo.com/docs/#read-balance)

           :param account_id: The unique identifier for the account which the balance belong to.
           :rtype: Dictionary representation of the current account balance.

        """
        url = "{0}/balance".format(self.API_URL)
        params = {'account_id': account_id}
        response = self.oauth_session.make_request(url, params=params)
        return response

    def get_webhooks(self, account_id:str) -> Dict[str, List]:
        """Gets the webhooks of a given account. (https://monzo.com/docs/#list-webhooks)

           :param account_id: The unique identifier for the account which the webhooks belong to.
           :rtype: A collection of webhooks that belong to an account.

        """
        url = "{0}/webhooks".format(self.API_URL)
        params = {'account_id': account_id}
        response = self.oauth_session.make_request(url, params=params)
        return response

    def get_first_webhook(self, account_id:str) -> Dict[str, Any]:
        """Gets the first webhook of a given account.

           :param account_id: The unique identifier for the account which the first webhook belong to.
           :rtype: A Dictionary representation of the first webhook belonging to an account, if it exists.
        """
        webhooks = self.get_webhooks(account_id)
        if len(webhooks['webhooks']) <= 0:
            raise LookupError('There are no webhooks associated with the account.')
        return webhooks['webhooks'][0]

    def delete_webhook(self, webhook_id) -> Dict:
        """Deletes the a specified webhook. (https://monzo.com/docs/#deleting-a-webhook)

           :param webhook_id: The unique identifier for the webhook to delete.
           :rtype: An empty Dictionary, if the deletion was successful.
        """
        url = "{0}/webhooks/{1}".format(self.API_URL, webhook_id)
        response = self.oauth_session.make_request(url, method='DELETE')
        return response

    def delete_all_webhooks(self) -> None:
        """Removes all webhooks associated with the first account, if it exists.

           :rtype: None
        """
        first_account = self.get_first_account()
        account_id = first_account['id']
        webhooks = self.get_webhooks(account_id)
        for webhook in webhooks['webhooks']:
            self.delete_webhook(webhook['id'])

    def register_webhook(self, webhook_url:str, account_id:str):
        """Registers a webhook. (https://monzo.com/docs/#registering-a-webhook)

           :param webhook_url: The webhook url to register.
           :param account_id: The unique identifier for the account to register the webhook.

           :rtype: Registers a webhook to an account.
        """
        url = "{0}/webhooks".format(self.API_URL)
        data = {'account_id': account_id, 'url': webhook_url}
        response = self.oauth_session.make_request(url, data=data)

    def register_attachment(self, transaction_id:str, file_url:str, file_type:str):
        """Attaches an image to a transaction. (https://monzo.com/docs/#register-attachment)

           :param transaction_id: The unique identifier for the transaction to register the attachment to.
           :param file_url: The url of the file to attach.
           :param transaction_id: The type of the file specified in file_url

           :rtype: Dictionary representation of the attachment that was just registered.
        """
        url = "{0}/attachment/register".format(self.API_URL)
        data = {'external_id': transaction_id,
              'file_url': file_url,
              'file_type': file_type}
        response = self.oauth_session.make_request(url,data=data)
        return response

    def deregister_attachment(self, attachment_id:str):
        """Removed a previously attached image from a transaction. (https://monzo.com/docs/#deregister-attachment)

            :param transaction_id: The unique identifier for the attachment to deregister.
            :rtype: An empty Dictionary, if the deregistration was successful.
        """
        url = "{0}/attachment/deregister".format(self.API_URL)
        data = {'id': attachment_id}
        response = self.oauth_session.make_request(url, data=data)
        return response


    def create_feed_item(self, account_id:str, feed_type:str, url:str, params:Dict):
        """Creates a feed item. (https://monzo.com/docs/#create-feed-item)

            :param account_id: The unique identifier for the account to create the feed item for.
            :param feed_type: The type of feed item (currently only `basic` is supported).
            :param url: The url to open if a feed item is tapped
            :param params: A map of parameters which vary based on type

            :rtype: An empty Dictionary, if the feed item creation was successful.
        """
        url = "{0}/feed".format(self.API_URL)
        data = {
            'account_id': account_id,
            'type': feed_type,
            'url': url,
            "params[title]": params['title'],
            "params[image_url]": params['image_url'],
            "params[body]": params['body']
        }
        response = self.request.post(url, data=data)
        return response


    def get_pots(self) -> Dict[str, List]:
        """Get all pots for a user. (https://monzo.com/docs/#list-pots)

           :rtype: A collection of pots for a user.

        """
        url = "{0}/pots".format(self.API_URL)
        response = self.oauth_session.make_request(url)
        return response


    def deposit_into_pot(self, pot_id:str, account_id:str, amount_in_pennies:int):
        """Move money from an account into a pot. (https://monzo.com/docs/#deposit-into-a-pot)

            :param pot_id: The unique identifier for the pot to deposit the money to.
            :param account_id: The unique identifier for the account to move the money from.
            :param amount_in_pennies: The amount of money to move to the pot in pennies.

            :rtype: A dictionary containing information on the pot that was updated.
        """
        url = "{0}/pots/{1}/deposit".format(self.API_URL, pot_id)
        unique_string = ''.join(random.choice(string.ascii_letters) for i in range(15))
        data = {
            'source_account_id': account_id,
            'amount': amount_in_pennies,
            'dedupe_id': unique_string
        }

        response = self.oauth_session.make_request(url, data=data)
        return response

    def withdraw_from_pot(self, account_id:str, pot_id:str, amount_in_pennies:str):
        """Move money from an account into a pot. (https://monzo.com/docs/#withdraw-from-a-pot)

            :param account_id: The unique identifier for the account to move the money to.
            :param pot_id: The unique identifier for the pot to withdraw the money from.
            :param amount_in_pennies: The amount of money to move to the pot in pennies.

            :rtype: A dictionary containing information on the pot that was updated.
        """
        url = "{0}/pots/{1}/withdraw".format(self.API_URL, pot_id)
        unique_string = ''.join(random.choice(string.ascii_letters) for i in range(15))
        data = {
            'destination_account_id': account_id,
            'amount': amount_in_pennies,
            'dedupe_id': unique_string
        }

        response = self.oauth_session.make_request(url, data=data, method='PUT')
        return response



from setuptools import setup

setup(name='monzo_tom_test',
      version='0.6.2',
      description='A python SDK for interacting with the Monzo API.',
      url='https://github.com/muyiwaolu/monzo-python',
      author='Muyiwa Olu-Ogunleye',
      author_email='m.oluogunleye94@gmail.com',
      license='MIT',
      packages=['monzo'],
      install_requires=[
          'requests',
          'requests-oauthlib',
          'python-dotenv'
      ],
      )
